import { Component } from 'react';
import { observe, unobserve, raw, isObservable } from '@nx-js/observer-util';
export { observable as store } from '@nx-js/observer-util';

var tasks = new Set();
var isStopped = false;

function add(task) {
  if (isStopped) {
    tasks.add(task);
  } else {
    runTask(task);
  }
}

function remove(task) {
  tasks.delete(task);
}

// this replaces the passed function with a function
// that batches all of its callback arguments
function batchCallbacks(fn) {
  return function batchedCallbacks() {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    var batchedArgs = args.map(function (arg) { return typeof arg === 'function' ? function () { return batch(arg); } : arg; });
    return fn.apply(this, batchedArgs);
  };
}

// this runs the passed function and delays all re-renders
// until the function is finished running
function batch(fn) {
  try {
    isStopped = true;
    return fn.apply(this, arguments);
  } finally {
    tasks.forEach(runTask);
    tasks.clear();
    isStopped = false;
  }
}

function runTask(task) {
  task();
}

// try to find the global object
// it is window in the DOM and global in NodeJS and React Native
var globalObj;
if (typeof window !== 'undefined') {
  globalObj = window;
} else if (typeof global !== 'undefined') {
  globalObj = global;
}

// do a sync batching for the most common task sources
// this should be removed when React's own batching is improved in the future
if (globalObj) {
  globalObj.setTimeout = batchCallbacks(globalObj.setTimeout);
  globalObj.setInterval = batchCallbacks(globalObj.setInterval);
  if (globalObj.requestAnimationFrame) {
    globalObj.requestAnimationFrame = batchCallbacks(globalObj.requestAnimationFrame);
  }
  if (globalObj.requestIdleCallback) {
    globalObj.requestIdleCallback = batchCallbacks(globalObj.requestIdleCallback);
  }
  // eslint-disable-next-line
  Promise.prototype.then = batchCallbacks(Promise.prototype.then);
  // eslint-disable-next-line
  Promise.prototype.catch = batchCallbacks(Promise.prototype.catch);
}

// DOM event handlers and HTTP event handlers don't have to be batched
// event handlers are batched by React
// HTTP event handlers are usually wrapped by Promises, which is covered above

var COMPONENT = Symbol('owner component');
var DUMMY_STATE = {};

function view(Comp, ref) {
  if ( ref === void 0 ) ref = {};
  var rawDevtool = ref.devtool;

  var isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);
  var BaseComp = isStatelessComp ? Component : Comp;

  var devtool = rawDevtool ? function (operation) { return rawDevtool(Object.assign({ Component: Comp }, operation)); } : undefined;

  // return a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
  // it decides when to run the new reactive methods and when to proxy to the original methods
  var ReactiveHOC = (function (BaseComp) {
    function ReactiveHOC(props, context) {
      var this$1 = this;

      BaseComp.call(this, props, context);

      this.state = this.state || {};
      this.state[COMPONENT] = this;

      // run a dummy setState to schedule a new render, avoid forceUpdate
      var updater = function () { return this$1.setState(DUMMY_STATE); };

      // create a reactive render for the component
      this.render = observe(this.render, {
        scheduler: {
          add: function () { return add(updater); },
          delete: function () { return remove(updater); }
        },
        debugger: devtool,
        lazy: true
      });
    }

    if ( BaseComp ) ReactiveHOC.__proto__ = BaseComp;
    ReactiveHOC.prototype = Object.create( BaseComp && BaseComp.prototype );
    ReactiveHOC.prototype.constructor = ReactiveHOC;

    ReactiveHOC.prototype.render = function render () {
      return isStatelessComp ? Comp(this.props, this.context) : BaseComp.prototype.render.call(this);
    };

    // react should trigger updates on prop changes, while easyState handles store changes
    ReactiveHOC.prototype.shouldComponentUpdate = function shouldComponentUpdate (nextProps, nextState) {
      var ref = this;
      var props = ref.props;
      var state = ref.state;

      // respect the case when user prohibits updates
      if (BaseComp.prototype.shouldComponentUpdate && !BaseComp.prototype.shouldComponentUpdate.call(this, nextProps, nextState)) {
        devtool && devtool({ type: 'render', renderType: 'blocked' });
        return false;
      }

      // return true if it is a reactive render or state changes
      if (state !== nextState) {
        devtool && devtool({ type: 'render', renderType: 'reactive' });
        return true;
      }

      // the component should update if any of its props shallowly changed value
      var keys = Object.keys(props);
      var nextKeys = Object.keys(nextProps);
      if (nextKeys.length !== keys.length || nextKeys.some(function (key) { return props[key] !== nextProps[key]; })) {
        devtool && devtool({
          type: 'render',
          renderType: 'normal',
          props: nextProps,
          oldProps: props
        });
        return true;
      }
      return false;
    };

    // add a custom deriveStoresFromProps lifecyle method
    ReactiveHOC.getDerivedStateFromProps = function getDerivedStateFromProps (props, state) {
      if (BaseComp.deriveStoresFromProps) {
        // inject all local stores and let the user mutate them directly
        var stores = mapStateToStores(state);
        BaseComp.deriveStoresFromProps.apply(this, [ props ].concat( stores ));
      }
      // respect user defined getDerivedStateFromProps
      if (BaseComp.getDerivedStateFromProps) {
        return BaseComp.getDerivedStateFromProps.call(this, props, state);
      }
      return null;
    };

    ReactiveHOC.prototype.componentWillUnmount = function componentWillUnmount () {
      // call user defined componentWillUnmount
      if (BaseComp.prototype.componentWillUnmount) {
        BaseComp.prototype.componentWillUnmount.call(this);
      }
      // clean up memory used by Easy State
      unobserve(this.render);
    };

    return ReactiveHOC;
  }(BaseComp));

  ReactiveHOC.displayName = Comp.displayName || Comp.name;
  // static props are inherited by class components,
  // but have to be copied for function components
  if (isStatelessComp) {
    for (var i = 0, list = Object.keys(Comp); i < list.length; i += 1) {
      var key = list[i];

      ReactiveHOC[key] = Comp[key];
    }
  }

  return ReactiveHOC;
}

function mapStateToStores(state) {
  // find store properties and map them to their none observable raw value
  // to do not trigger none static this.setState calls
  // from the static getDerivedStateFromProps lifecycle method
  var component = state[COMPONENT];
  return Object.keys(component).map(function (key) { return component[key]; }).filter(isObservable).map(raw);
}

export { view, batch };
